//
//	Copyright  Nemetschek Vectorworks, Inc.
//	Use of this file is governed by the Nemetschek Vectorworks SDK License Agreement
//	http://developer.vectorworks.net/index.php?title=Vectorworks_SDK_License
//

#pragma once

namespace VWFC
{
	namespace Math
	{
		using namespace VectorWorks::Math;

		// ------------------------------------------------------------------------------------
		// polygonization default (recommended) values.
		const short			kPolygonize_MaxRecursLevel		= 8;
		const double		kPolygonize_ArcResolution		= 0.5;		// 0.5 degrees.
		const size_t		kPolygonize_MaxIterCount		= 5;
		const double		kPolygonize_IntersEpsilon		= 0.0001;

		// ------------------------------------------------------------------------------------
		class VWPolygon2D
		{
		public:
							VWPolygon2D();
							VWPolygon2D(const TSTLPoints2DArray& src);
							VWPolygon2D(const TSTLPoints2DVWArray& src);
							VWPolygon2D(IPolyDefPtr pPolygonDef);
							VWPolygon2D(const VWPolygon2D& src);
			virtual			~VWPolygon2D();

			VWPolygon2D&	operator=(const TSTLPoints2DArray& src);
			VWPolygon2D&	operator=(const TSTLPoints2DVWArray& src);
			VWPolygon2D&	operator=(const VWPolygon2D& src);
			VWPolygon2D&	operator=(IPolyDefPtr pPolygonDef);

		// operations
		public:
			// setting polygon
			void			SetPolygon(const TSTLPoints2DArray& src);
			void			SetPolygon(const TSTLPoints2DVWArray& src);
			void			SetPolygon(IPolyDefPtr pPolygonDef);
			void			GetPolygon(TSTLPoints2DArray& out) const;
			void			GetPolygon(TSTLPoints2DVWArray& out) const;
			void			GetPolygon(IPolyDefPtr pPolygonDef) const;

			// vertex operations
			void			ClearVertices();
			void			ClearVertex(size_t at);
			void			AddVertex(double x, double y, EVertexType type=eVertexType_Corner, double dArcRadius=0);
			void			AddVertex(const VWPoint2D& pt, EVertexType type=eVertexType_Corner, double dArcRadius=0);
			void			AddVertex(const VWPolygon2D& poly);
			void			InsertVertex(size_t at, double x, double y, EVertexType type=eVertexType_Corner, double dArcRadius=0);
			void			InsertVertex(size_t at, const VWPoint2D& pt, EVertexType type=eVertexType_Corner, double dArcRadius=0);
			void			InsertVertex(size_t at, const VWPolygon2D& poly);

			size_t			GetVertexCount() const;
			bool			IsEmpty() const;

			VWPoint2D&		GetVertexAt(size_t at, bool bLoopVerts=false);
			const VWPoint2D& GetVertexAt(size_t at, bool bLoopVerts=false) const;
			EVertexType		GetVertexType(size_t at) const;
			void			SetVertexType(size_t at, EVertexType type);
			double			GetArcVertexRadius(size_t at) const;
			void			SetArcVertexRadius(size_t at, double dRadius);
			bool			GetVertexVisible(size_t at) const;
			void			SetVertexVisible(size_t at, bool bVisible);
			// return a line from (at, at+1) vertex (at the end the line is (len-1, 0)
			VWLine2D		GetLineAt(size_t at) const;

			// modify closed flag
			void			SetClosed(bool isClosed);
			bool			IsClosed() const;
			void			SetAllVertexVisible(bool bVisible);
			
			// get bounding box of this polygon
			void			GetPolygonBounds(double& outMinX, double& outMaxX, double& outMinY, double& outMaxY) const;
			VWRectangle2D	GetPolygonBounds() const;
			VWPoint2D		GetPolygonBoundsCenter() const;
			// get the bouding box which is parallel to given line
			// the result polygon contains 4 vertices of the bounding box
			void			GetPolygonBounds(const VWLine2D& line, VWPolygon2D& outBounds) const;

			// get the bounding box which is parallel to the given direction (a similar approach).
			// the result is 4 vertices (UL, UR, LL, LR) of the bounding box
			// XXX Bobi: in contrast to the previous, this one first polygonizes 
			// the polygon and then gets the bounding box. It works for arbitrary polylines.
			void			GetPolygonBounds(const VWPoint2D& vDirection, VWPoint2D& vUL, VWPoint2D& vUR, VWPoint2D& vLL, VWPoint2D& vLR) const;


			// reverse the order of the points (if CW to CCW or CCW to CW)
			void			ReversePoints(const bool& kbDoFullReversion = false);

			// determine if the polygon is Clock Wise oriented
			bool			IsClockWise() const;

			// return true if the passed polygon has more than 3 different vertices
			bool			IsValidPolygon(double dEpsilon) const;

			// removed all same points in the polygon
			void			CleanUpPolygon(double dEpsilon);

			// This function works as follows depending on 'doClosenessCheck':
			// ( doClosenessCheck == false )	- clean up all polygon vertices which are collinear with their neighbors.
			// ( doClosenessCheck == true  )	- clean up all polygon vertices which are collinear with their neighbors and are too close to their neighbors.
			// Note: works on polygon with only corner vertices.
			void			CleanUp_FeaturePreserving(bool doClosenessCheck, double epsCollinear, double epsCloseness);

			// check polygon for self intersections. return true is so.
			// 'doConsiderTouching' tells if touching will be considered as intersection.
			bool			IsSelfIntersecting(bool doConsiderTouching = true) const;

		// mathematics
		public:
			// join this polygon to the given polygon only if both polygons are not closed and have touching endpoints (within epsilon).
			bool			AddAdjoiningPolygon(const VWPolygon2D& adjoiningPoly, double epsilon, bool checkTypes = false);

			// find indices of the first 3 'apposite' polygon points (not too close and not collinear).
			bool			GetAppositePtIndices(size_t& ptIndex0, size_t& ptIndex1, size_t& ptIndex2) const;
			// find the first 3 'apposite' polygon points (not too close and not collinear).
			bool			GetThreeAppositePoints(VWPoint2D& pt0, VWPoint2D& pt1, VWPoint2D& pt2) const;

			// find index of next 'apposite' polygon point (not too close).
			bool			GetNextAppositePtIndex(size_t inPtIndex, size_t& outNextPtIndex, double dEpsilon) const;
			// find next 'apposite' polygon point (not too close).
			bool			GetNextAppositePt(size_t inPtIndex, VWPoint2D& outNextPt, double dEpsilon) const;
			// find index of previous 'apposite' polygon point (not too close).
			bool			GetPrevAppositePtIndex(size_t inPtIndex, size_t& outPrevPtIndex, double dEpsilon) const;
			// find previous 'apposite' polygon point (not too close).
			bool			GetPrevAppositePt(size_t inPtIndex, VWPoint2D& outPrevPt, double dEpsilon) const;

			// transform the polygon with specified matrix
			void			Transform(const VWTransformMatrix& mat);
			// move the polygon by the given values
			void			MovePolygon(double cx, double cy);
			// rotate the polygon by the center of the bounding box
			void			Rotate(double degAngle);
			// rotate by specified center
			void			Rotate(const VWPoint2D& pt, double degAngle);

			// calculate the area of the polygon
			// positive if polygon is CCW -- negative if CW
			double			GetArea() const;
			// calculate the perimeter of the polygon (sum of distances between the vertices)
			double			GetPerimeter() const;
			// calculate the length of the curve formed by the polyline
			// Does not include visibility or closed state!
			double			GetLength() const;
			// calculate the length of the curve formed by the polyline up to the specified point
			// if the point is not on the poly 
			// Does not include visibility or closed state!
			double			GetLength(const VWPoint2D& pt, double dEpsilon) const;
			double			GetLength(size_t fromVert, size_t toVert) const;
			// result is a point along the poly on the specified offset
			bool			GetPtOnPoly(double offset, VWPoint2D& outPt) const;
			bool			GetPtOnPoly(double offset, VWPoint2D& outPt, VWPoint2D& outNormal) const;
			// finds the closest point on poly of the given point.
			VWPoint2D		GetClosestPtOnPoly(const VWPoint2D& pt, double& outDist, int& outOnSegIndex) const;
			// finds the closest projection point on the polygon of the given point (if exists). Note: works on polygon with only corner vertices.
			// returns closest projection point, squared distance to the projection point and index of the segment where the projection point lies.
			bool			GetClosestProjPtOnPoly_Fast(const VWPoint2D& pt, double& outSqrDist, int& outOnSegIndex, VWPoint2D& outProjPt) const;

			// calculates an offset ring poly ouside or inside of this poly.
			void			GetOffsetRingPoly(double offset, bool doOutsideOffset, VWPolygon2D& offsetRingPoly) const;

			// get polygonization:
			// has parameters to control the density of the resulting polygon. recommended to use defaults !
			// if 'interpolate' is true, then bezier and cubic segments will be interpolated by adding cubic vertices, ...
			// ... so result will be interpolation poly with only corner and cubic vertices.
			void			GetPolygonized(	VWPolygon2D& outPolygon,
											bool	interpolate		= false,
											short	splineRecusLevel= kPolygonize_MaxRecursLevel,
											double	arcDegRes		= kPolygonize_ArcResolution,
											size_t	iterationCount	= kPolygonize_MaxIterCount) const;

			// check if it is an equalish polygon
			bool			EqualishTo(const VWPolygon2D& poly) const;

			// is specified point on the edge of the specified polygon
			bool			IsPtOnPoly2D(double ptX, double ptY, double dEpsilon) const;
			bool			IsPtOnPoly2D(const VWPoint2D& pt, double dEpsilon) const;
			bool			IsPtOnPoly2D_Fast(const VWPoint2D& inPt, int& outOnSegIndex, double epsilon) const;

			// determines if a point is in the area of a given polygon (the 2D projection)
			bool			IsPtInPoly2D(double ptX, double ptY, double dEpsilon) const;
			bool			IsPtInPoly2D(const VWPoint2D& pt, double dEpsilon) const;
			bool			IsPtInPoly2D_Fast(double ptX, double ptY, double dEpsilon) const;

			// combined operations
			bool			IsPtInOrOnPoly2D(double ptX, double ptY, double dEpsilon) const;
			bool			IsPtInOrOnPoly2D(const VWPoint2D& pt, double dEpsilon) const;

			// determines if a points is inside and if it is on the poly
			void			IsPtOnPoly2D(double ptX, double ptY, bool& outOnPoly, bool& outInPoly, double dEpsilon) const;
			void			IsPtOnPoly2D(const VWPoint2D& pt, bool& outOnPoly, bool& outInPoly, double dEpsilon) const;

			// return true if the specified line intersects the poly
			bool			IsIntersecting(const VWLine2D& line, double dEpsilon, VWPoint2D* /*out*/ pInsertPt=NULL ) const;
			bool			IsIntersecting(const VWLine2D& line, double dEpsilon, TSTLPoints2DArray& outArrIntersectPts ) const;
			bool			IsIntersecting(const VWPoint2D& a, const VWPoint2D& b, double dEpsilon, VWPoint2D* /*out*/ pInsertPt=NULL ) const;
			bool			IsIntersecting(const VWPoint2D& a, const VWPoint2D& b, double dEpsilon, TSTLPoints2DArray& outArrIntersectPts ) const;

			// return true if the given polygon and this polygon intersects each other (including touching)
			// intersect 'true' means that they have intersection points or same points
			// 'outHaveFullyOutsideVerts' true when the specified 'poly' have fully outside vertices from this poly
			bool			IsIntersecting(const VWPolygon2D& poly, double dEpsilon) const;

			// return true if part of a polygon is inside a vicinity of this polygon.
			// this means that polygons intersect or close enough.
			bool			IsInVicinity(const VWPolygon2D& poly, double dVicinity) const;

			// retun true if the given polygon and this polygon intersects each other (including touching)
			// outIsTouching returns if they are tourching
			bool			IsIntersecting(const VWPolygon2D& poly, bool& outIsTouching, double dEpsilon) const;
			bool			IsIntersecting(const VWPolygon2D& poly, bool& outIsTouching, VWPoint2D& outIntersPt, double dEpsilon) const;

			// returns true if the given polygon is fully within this polygon
			bool			IsFullyInside(const VWPolygon2D& poly, double dEpsilon) const;
			// returns true if the given polygon is fully outside this polygon
			bool			IsFullyOutside(const VWPolygon2D& poly, double dEpsilon) const;

			// returns true if the given polygon is within this polygon, including having common points
			bool			IsInside(const VWPolygon2D& poly, double dEpsilon) const;
			// returns true if the given polygon is fully outside this polygon, including having common points
			bool			IsOutside(const VWPolygon2D& poly, double dEpsilon) const;
			
			// get geometry center of the polygon
			VWPoint2D		GetCenter() const;
			// get an inside point: prefer points that are not too close to the polygon boundary.
			bool			GetPtInsidePoly(VWPoint2D& insidePt, double dEpsilon) const;
			// get random point that lies in the poly
			// XXX Bobi: not advisable to use: for polygon with small area it could take very long time to finish !
			VWPoint2D		GetRandomPtInPoly(double dEpsilon) const;

			// get bounding points along line
			// the result points are projection on the line of the two points from the poly wich are bound points along the line
			void			GetBoundingProjOnLine(VWPoint2D&/*in-out*/ minProjPt, VWPoint2D&/*in-out*/ maxProjPt, double dEpsilon) const;

			// get an inside point near the vertex at
			VWPoint2D		GetInsidePtNearVertexAt(size_t at);
			// get set of inside points near polygon's vertices
			TSTLPoints2DArray	GetInsideNearPts();

			// transform polygon's points
			void			PointsTransform(const VWTransformMatrix& matrix);
			void			InversePointsTransform(const VWTransformMatrix& matrix);

		// Math tools
		public:
			// is pt (x0,y0) is inside or/and on the specified triangle (x1,y1):(x2,y2):(x3,y3)
			static void		IsPtInTriangle( double x0, double y0,
											double x1, double y1,
											double x2, double y2,
											double x3, double y3,
											bool& outInTri, bool& outOnTri);
			static void		IsPtInTriangle(	double x0, double y0,
											double x1, double y1,
											double x2, double y2,
											double x3, double y3,
											bool& outInTri,
											bool& outOnTri,
											double epsilon);
			static void		IsPtInTriangle(	double x0, double y0,
											double x1, double y1,
											double x2, double y2,
											double x3, double y3,
											bool& outInTri,
											int& outOnSegIndex,
											int& outOnPtIndex,
											double epsilon);
			// is pt (x0,y0) on the border of the specified triangle (x1,y1):(x2,y2):(x3,y3)
			static void		IsPtOnTriangleBorder( double x0, double y0,
												  double x1, double y1,
												  double x2, double y2,
												  double x3, double y3,
												  int& outOnSegIndex,
												  int& outOnPtIndex,
												  double epsilon);
			// faster check if two polygons intersect each other. NOTE: no epsilon is used.
			static bool		ArePolygonsIntersecting_Fast(const VWPolygon2D& polyA, const VWPolygon2D& polyB, bool& outNoCommonArea);

		public:
			class CLockForEdit
			{
			public:
							CLockForEdit(VWPolygon2D& poly);
				virtual		~CLockForEdit();

			private:
				VWPolygon2D&	fPoly;
			};

		// internal funcs
		protected:
			void			RecalculateBoundingBox();

		protected:
			struct SVertex
			{
									SVertex();
									SVertex(const SVertex& src);
				virtual				~SVertex();

				// operators.
				SVertex&			operator=(const SVertex& src);
				bool				operator==(const SVertex& that) const;
				bool				operator!=(const SVertex& that) const;

				// data.
				EVertexType			type;
				VWPoint2D			pt;
				double				dArcRadius;
				bool				bVisible;
			};
			typedef	std::vector<SVertex>	TVertexArray;

			struct SExtVertex : public SVertex
			{
									SExtVertex();
									SExtVertex(const SExtVertex& src);
									SExtVertex(const SVertex& src);
				virtual				~SExtVertex();

				// operators.
				SExtVertex&			operator=(const SExtVertex& src);
				SExtVertex&			operator=(const SVertex& src);
				bool				operator==(const SExtVertex& that) const;
				bool				operator!=(const SExtVertex& that) const;

				// data.
				VWPoint2D			controlPt;
			};
			typedef	std::vector<SExtVertex>	TExtVertexArray;

		// polygonization utilities.
		protected:
			void			CalculatePolyline(	TExtVertexArray& arrVertices,
												size_t iterationCount = kPolygonize_MaxIterCount) const;

			void			SwapControlPts(TExtVertexArray& arrVertices) const;

			void			PolygonizeSection(	const SExtVertex& lastVert, const SExtVertex& thisVert, const SExtVertex& nextVert,
												VWPolygon2D& outPolygon,
												bool	interpolate		= false,
												short	splineRecusLevel= kPolygonize_MaxRecursLevel,
												double	arcDegRes		= kPolygonize_ArcResolution) const;
			void			PolygonizeSplineSubSection(const VWPoint2D& v1, const VWPoint2D& v2, const VWPoint2D& v3, short level, bool interpolate, VWPolygon2D& outPolygon) const;

			// arc polygonization utilities.
			bool			CalcChamfer(const VWPoint2D& pt1, const VWPoint2D& int0Value, const VWPoint2D& pt4, VWPoint2D& lastCam, VWPoint2D& thisCam, VWPoint2D& center, double& radius, double& lastAngle, double& thisAngle) const;
			VWPoint2D		NextLoc(double a, const VWPoint2D& cen, double radius) const;

			void			PolygonizeArc(const VWPoint2D& lastPt, const VWPoint2D& thisPt, const VWPoint2D& nextPt, VWPolygon2D& outPolygon) const;
			void			PolygonizeArc(const VWPoint2D &lastPt, const VWPoint2D& nextPt, double radius, VWPoint2D& lastCam, VWPoint2D& thisCam, VWPoint2D& center, double theStart, double theArc, VWPolygon2D& outPolygon) const;
			void			PolygonizeArcSegment(double inStart, double inArc, const VWPoint2D& cen, double radius, VWPolygon2D& outPolygon) const;
			void			PolygonizeArcSegment(double arcStart, double arcEnd, short currentLevel, const VWPoint2D& cen, double radius, VWPolygon2D& outPolygon) const;

		protected:
			bool									fLockedForEdit;
			double									fdMinX;
			double									fdMaxX;
			double									fdMinY;
			double									fdMaxY;
			TVertexArray							farrVertices;
			bool									fbIsClosed;
		};

		// array of polygon defintion
		typedef		std::vector<VWPolygon2D>									TSTLPolygons2DArray;
		typedef		std::vector<VWPolygon2D, std::vwallocator<VWPolygon2D> >	TSTLPolygons2DVWArray;
	}
}