//
//	Copyright  Nemetschek Vectorworks, Inc.
//	Use of this file is governed by the Nemetschek Vectorworks SDK License Agreement
//	http://developer.vectorworks.net/index.php?title=Vectorworks_SDK_License
//

#include "StdHeaders.h"
#include "../VWFCLibrary.h" // Use for Intellisense support during dev; remove before committing

using namespace VectorWorks;
using namespace VWFC::VWObjects;

const OSType kDesignLayerViewportDataTypeTag = 'DLVT'; // Design Layer viewport data

////////////////////////////////////////////////////////////////////////////////
// Constructors and Destructor
////////////////////////////////////////////////////////////////////////////////

/* Creates a VWViewportObj associated with the provided handle
 * Throws an exception if the handle is the wrong type
 */
VWViewportObj::VWViewportObj(MCObjectHandle hViewport)
{
	VWFC_ASSERT( VWViewportObj::IsViewportObject(hViewport) || VWViewportObj::IsDesignLayerViewportObject(hViewport) );
	if (!VWViewportObj::IsViewportObject(hViewport) && !VWViewportObj::IsDesignLayerViewportObject(hViewport))
		THROW_VWFC_EXCEPTION( VWCannotCreateObjectException, 0, "bad handle type when creating" );
	fhObject = hViewport;
}

/* Creates a new viewport and returns the associated VWViewportObj
 */
bool VWViewportObj::CreateViewport(MCObjectHandle parentHandle, VWViewportObj& outViewportObj)
{
	MCObjectHandle viewportHandle = gSDK->CreateViewport(parentHandle);
	if (viewportHandle)
	{
		outViewportObj = VWViewportObj(viewportHandle);
		return true;
	}
	else {
		return false;
	}
}

////////////////////////////////////////////////////////////////////////////////
// Operators
////////////////////////////////////////////////////////////////////////////////

/* Implements the '=' operator for assignment
 */
VWViewportObj& VWViewportObj::operator=(const VWViewportObj& src)
{
	fhObject = src.fhObject;
	return *this;
}

////////////////////////////////////////////////////////////////////////////////
// Document Manipulators
////////////////////////////////////////////////////////////////////////////////

void VWViewportObj::Update()
{
	gSDK->UpdateViewport(fhObject);
}

bool VWViewportObj::AddAnnotationObject(MCObjectHandle annotationHandle)
{
	return gSDK->AddViewportAnnotationObject(fhObject, annotationHandle);
}

bool VWViewportObj::HasCropObject() const
{
	return gSDK->ViewportHasCropObject(fhObject);
}

bool VWViewportObj::GetCropObject(MCObjectHandle& outCropHandle) const
{
	outCropHandle = gSDK->GetViewportCropObject(fhObject);
	return (outCropHandle != nil);
}

bool VWViewportObj::SetCropObject(MCObjectHandle inCropHandle)
{
	return gSDK->SetViewportCropObject(fhObject, inCropHandle);
}

bool VWViewportObj::GetCropVisible() const
{
	Boolean cropVisible;
	TVariableBlock block(cropVisible);
	gSDK->GetObjectVariable(fhObject, ovViewportCropVisible, block);
	block.GetBoolean(cropVisible);
	
	return cropVisible;
}

void VWViewportObj::SetCropVisible(bool cropVisible)
{
	TVariableBlock block((Boolean) cropVisible);
	gSDK->SetObjectVariable(fhObject, ovViewportCropVisible, block);
}

bool VWViewportObj::GetGroup(short groupType, MCObjectHandle& outGroupHandle) const
{
	outGroupHandle = gSDK->GetViewportGroup(fhObject, groupType);
	return (outGroupHandle != nil);
}

WorldPt VWViewportObj::GetPosition() const
{
	Real64 posX, posY;
	TVariableBlock block(posX);
	gSDK->GetObjectVariable(fhObject, ovViewportXPosition, block);
	block.GetReal64(posX);
	gSDK->GetObjectVariable(fhObject, ovViewportYPosition, block);
	block.GetReal64(posY);
	
	return WorldPt(posX, posY);
}

/* The angle the viewport subtends with the X-axis in degrees
 */
Real64 VWViewportObj::GetAngleToX() const
{
	Real64 angleX;
	TVariableBlock block(angleX);
	gSDK->GetObjectVariable(fhObject, ovViewportAngleWithXAxis, block);
	block.GetReal64(angleX);
	
	return angleX;
}

TXString VWViewportObj::GetDescription() const
{
	Str255 description;
	TVariableBlock block(description);
	gSDK->GetObjectVariable(fhObject, ovViewportDescription, block);
	block.GetStr255(description);

	return (TXString) description;
}

void VWViewportObj::SetDescription(TXString description)
{
	Str255 desc;
	description.Truncate(63);
	description.CopyInto(desc);
	TVariableBlock block(desc);
	gSDK->SetObjectVariable(fhObject, ovViewportDescription, block);
}

TXString VWViewportObj::GetLocator() const
{
	Str255 locator;
	TVariableBlock block(locator);
	gSDK->GetObjectVariable(fhObject, ovViewportLocator, block);
	block.GetStr255(locator);

	return (TXString) locator;
}

void VWViewportObj::SetLocator(TXString locator)
{
	Str255 loc;
	locator.Truncate(63);
	locator.CopyInto(loc);
	TVariableBlock block(loc);
	gSDK->SetObjectVariable(fhObject, ovViewportLocator, block);
}

bool VWViewportObj::IsLinked() const
{
	Boolean isLinked;
	TVariableBlock block(isLinked);
	gSDK->GetObjectVariable(fhObject, ovViewportIsLinked, block);
	block.GetBoolean(isLinked);

	return isLinked;
}

bool VWViewportObj::GetLayerHeightIgnored() const
{
	Boolean layerHeightIgnored;
	TVariableBlock block(layerHeightIgnored);
	gSDK->GetObjectVariable(fhObject, ovViewportLayerHeightIgnored, block);
	block.GetBoolean(layerHeightIgnored);

	return layerHeightIgnored;
}

////////////////////////////////////////////////////////////////////////////////
// View
////////////////////////////////////////////////////////////////////////////////

EProjectionType VWViewportObj::GetProjectionType() const
{
	Sint16 projectionType;
	TVariableBlock block(projectionType);
	gSDK->GetObjectVariable(fhObject, ovViewportProjectionType, block);
	block.GetSint16(projectionType);
	
	return (EProjectionType) projectionType;
}

void VWViewportObj::SetProjectionType(EProjectionType projectionType)
{
	TVariableBlock block((Sint16) projectionType);
	gSDK->SetObjectVariable(fhObject, ovViewportProjectionType, block);
}

Real64 VWViewportObj::GetPerspectiveDistance() const
{
	Real64 distance;
	TVariableBlock block(distance);
	gSDK->GetObjectVariable(fhObject, ovViewportPerspectiveDistance, block);
	block.GetReal64(distance);

	return distance;
}

void VWViewportObj::SetPerspectiveDistance(Real64 distance)
{
	TVariableBlock block(distance);
	gSDK->SetObjectVariable(fhObject, ovViewportPerspectiveDistance, block);
}

bool VWViewportObj::GetProject2D() const
{
	Boolean project2D;
	TVariableBlock block(project2D);
	gSDK->GetObjectVariable(fhObject, ovViewportProject2D, block);
	block.GetBoolean(project2D);
	return project2D;
}

void VWViewportObj::SetProject2D(bool project2D)
{
	TVariableBlock block((Boolean) project2D);
	gSDK->SetObjectVariable(fhObject, ovViewportProject2D, block);
}

EViewType VWViewportObj::GetViewType() const
{
	short viewType;
	TVariableBlock block(viewType);
	gSDK->GetObjectVariable(fhObject, ovViewportViewType, block);
	block.GetSint16(viewType);
	return (EViewType) viewType;
}

void VWViewportObj::SetViewType(EViewType viewType)
{
	TVariableBlock block((short) viewType);
	gSDK->SetObjectVariable(fhObject, ovViewportViewType, block);
}

bool VWViewportObj::HasCamera() const
{
	Boolean hasCamera;
	TVariableBlock block(hasCamera);
	gSDK->GetObjectVariable(fhObject, ovViewportHasCamera, block);
	block.GetBoolean(hasCamera);
	return hasCamera;
}

////////////////////////////////////////////////////////////////////////////////
// Rendering
////////////////////////////////////////////////////////////////////////////////

ERenderType VWViewportObj::GetRenderType() const
{
	Sint16 renderType;
	TVariableBlock block(renderType);
	gSDK->GetObjectVariable(fhObject, ovViewportRenderType, block);
	block.GetSint16(renderType);

	return (ERenderType) renderType;
}

void VWViewportObj::SetRenderType(ERenderType renderType)
{
	TVariableBlock block((Sint16) renderType);
	gSDK->SetObjectVariable(fhObject, ovViewportRenderType, block);
}

ERenderType VWViewportObj::GetForegroundRenderType() const
{
	Sint16 renderType;
	TVariableBlock block(renderType);
	gSDK->GetObjectVariable(fhObject, ovViewportForegroundRenderType, block);
	block.GetSint16(renderType);

	return (ERenderType) renderType;
}

void VWViewportObj::SetForegroundRenderType(ERenderType renderType)
{
	TVariableBlock block((Sint16) renderType);
	gSDK->SetObjectVariable(fhObject, ovViewportForegroundRenderType, block);
}

/* Returns whether the viewport is out of date (dirty)
 */
bool VWViewportObj::IsDirty() const
{
	Boolean isDirty;
	TVariableBlock block(isDirty);
	gSDK->GetObjectVariable(fhObject, ovViewportDirty, block);
	block.GetBoolean(isDirty);
	return isDirty;
}

InternalIndex VWViewportObj::GetRenderBackground() const
{
	InternalIndex background;
	TVariableBlock block((long) background);
	gSDK->GetObjectVariable(fhObject, ovViewportRenderBackground, block);
	block.GetSint32(background);
	return background;
}

void VWViewportObj::SetRenderBackground(InternalIndex background)
{
	TVariableBlock block((long) background);
	gSDK->SetObjectVariable(fhObject, ovViewportRenderBackground, block);
}

bool VWViewportObj::GetAmbientIsOn() const
{
	Boolean isOn;
	TVariableBlock block(isOn);
	gSDK->GetObjectVariable(fhObject, ovViewportAmbientIsOn, block);
	block.GetBoolean(isOn);
	return isOn;
}

void VWViewportObj::SetAmbientIsOn(bool isOn)
{
	TVariableBlock block((Boolean) isOn);
	gSDK->SetObjectVariable(fhObject, ovViewportAmbientIsOn, block);
}

Real64 VWViewportObj::GetAmbientBrightness() const
{
	Real64 brightness;
	TVariableBlock block(brightness);
	gSDK->GetObjectVariable(fhObject, ovViewportAmbientBrightness, block);
	block.GetReal64(brightness);
	return brightness;
}

void VWViewportObj::SetAmbientBrightness(Real64 brightness)
{
	TVariableBlock block(brightness);
	gSDK->SetObjectVariable(fhObject, ovViewportAmbientBrightness, block);
}

WorldPt3 VWViewportObj::GetAmbientColor() const
{
	WorldPt3 color;
	TVariableBlock block(color);
	gSDK->GetObjectVariable(fhObject, ovViewportAmbientColor, block);
	block.GetWorldPt3(color);
	return color;
}

void VWViewportObj::SetAmbientColor(WorldPt3 color)
{
	TVariableBlock block(color);
	gSDK->SetObjectVariable(fhObject, ovViewportAmbientColor, block);
}

InternalIndex VWViewportObj::GetHDRIBackground() const
{
	InternalIndex background;
	TVariableBlock block((long) background);
	gSDK->GetObjectVariable(fhObject, ovViewportHDRIBackground, block);
	block.GetSint32(background);
	return background;
}

bool VWViewportObj::GetShowWalls() const
{
	Boolean showWalls;
	TVariableBlock block(showWalls);
	gSDK->GetObjectVariable(fhObject, ovViewportShowWallComponents, block);
	block.GetBoolean(showWalls);
	return showWalls;
}

void VWViewportObj::SetShowWalls(bool showWalls)
{
	TVariableBlock block((Boolean) showWalls);
	gSDK->SetObjectVariable(fhObject, ovViewportShowWallComponents, block);
}

bool VWViewportObj::GetGrayAsTransparent() const
{
	Boolean grayAsTransparent;
	TVariableBlock block(grayAsTransparent);
	gSDK->GetObjectVariable(fhObject, ovViewportGrayTransparent, block);
	block.GetBoolean(grayAsTransparent);
	return grayAsTransparent;
}

void VWViewportObj::SetGrayAsTransparent(bool grayAsTransparent)
{
	TVariableBlock block((Boolean) grayAsTransparent);
	gSDK->SetObjectVariable(fhObject, ovViewportGrayTransparent, block);
}

bool VWViewportObj::GetFlipText() const
{
	Boolean flipText;
	TVariableBlock block(flipText);
	gSDK->GetObjectVariable(fhObject, ovViewportFlipText, block);
	block.GetBoolean(flipText);
	return flipText;
}

void VWViewportObj::SetFlipText(bool flipText)
{
	TVariableBlock block((Boolean) flipText);
	gSDK->SetObjectVariable(fhObject, ovViewportFlipText, block);
}

bool VWViewportObj::GetBlackAndWhite() const
{
	Boolean blackAndWhite;
	TVariableBlock block(blackAndWhite);
	gSDK->GetObjectVariable(fhObject, ovViewportBlackAndWhite, block);
	block.GetBoolean(blackAndWhite);
	return blackAndWhite;
}

void VWViewportObj::SetBlackAndWhite(bool blackAndWhite)
{
	TVariableBlock block((Boolean) blackAndWhite);
	gSDK->SetObjectVariable(fhObject, ovViewportBlackAndWhite, block);
}

bool VWViewportObj::GetDisplayPlanar() const
{
	Boolean displayPlanar;
	TVariableBlock block(displayPlanar);
	gSDK->GetObjectVariable(fhObject, ovViewportDisplayPlanar, block);
	block.GetBoolean(displayPlanar);
	return displayPlanar;
}

void VWViewportObj::SetDisplayPlanar(bool displayPlanar)
{
	TVariableBlock block((Boolean) displayPlanar);
	gSDK->SetObjectVariable(fhObject, ovViewportDisplayPlanar, block);
}

////////////////////////////////////////////////////////////////////////////////
// Section Classes
////////////////////////////////////////////////////////////////////////////////

/* Gets the class which provides attributes for the overall 2D cross-section, 
 * or just the cross-section of structural elements (see GetSection2DOtherClass below)
 */
InternalIndex VWViewportObj::GetSection2DClass() const
{
	InternalIndex classIndex;
	TVariableBlock block((long) classIndex);
	gSDK->GetObjectVariable(fhObject, ovSectionVP2DAttributeClass, block);
	block.GetSint32(classIndex);
	return classIndex;
}

/* Gets the class which provides attributes for the 2D cross-sections
 * of non-structural elements
 */
InternalIndex VWViewportObj::GetSection2DOtherClass() const
{
	InternalIndex classIndex;
	TVariableBlock block((long) classIndex);
	gSDK->GetObjectVariable(fhObject, ovSectionVP2DOtherAttribClass, block);
	block.GetSint32(classIndex);
	return classIndex;
}

/* Gets the class which provides the fill style of objects beyond the section plane
 */
InternalIndex VWViewportObj::GetSection3DFillClass() const
{
	InternalIndex classIndex;
	TVariableBlock block((long) classIndex);
	gSDK->GetObjectVariable(fhObject, ovSectionVP3DFillStyleClass, block);
	block.GetSint32(classIndex);
	return classIndex;
}

/* Gets the class which provides the line style of objects beyond the section plane
 */
InternalIndex VWViewportObj::GetSection3DLineClass() const
{
	InternalIndex classIndex;
	TVariableBlock block((long) classIndex);
	gSDK->GetObjectVariable(fhObject, ovSectionVP3DLineStyleClass, block);
	block.GetSint32(classIndex);
	return classIndex;
}

////////////////////////////////////////////////////////////////////////////////
// Scale
////////////////////////////////////////////////////////////////////////////////

Real64 VWViewportObj::GetScale() const
{
	Real64 scale;
	TVariableBlock block(scale);
	gSDK->GetObjectVariable(fhObject, ovViewportScale, block);
	block.GetReal64(scale);

	return scale;
}

void VWViewportObj::SetScale(Real64 scale)
{
	TVariableBlock block(scale);
	gSDK->SetObjectVariable(fhObject, ovViewportScale, block);
}

Real64 VWViewportObj::GetLineWeightScale() const
{
	Real64 scale;
	TVariableBlock block(scale);
	gSDK->GetObjectVariable(fhObject, ovViewportLineWeightScale, block);
	block.GetReal64(scale);

	return scale;
}

void VWViewportObj::SetLineWeightScale(Real64 scale)
{
	TVariableBlock block(scale);
	gSDK->SetObjectVariable(fhObject, ovViewportLineWeightScale, block);
}

Real64 VWViewportObj::GetArrowheadScale() const
{
	Real64 scale;
	TVariableBlock block(scale);
	gSDK->GetObjectVariable(fhObject, ovViewportArrowheadScale, block);
	block.GetReal64(scale);

	return scale;
}

void VWViewportObj::SetArrowheadScale(Real64 scale)
{
	TVariableBlock block(scale);
	gSDK->SetObjectVariable(fhObject, ovViewportArrowheadScale, block);
}

Real64 VWViewportObj::GetDashedLineScale() const
{
	Real64 scale;
	TVariableBlock block(scale);
	gSDK->GetObjectVariable(fhObject, ovViewportDashedLineScale, block);
	block.GetReal64(scale);

	return scale;
}

void VWViewportObj::SetDashedLineScale(Real64 scale)
{
	TVariableBlock block(scale);
	gSDK->SetObjectVariable(fhObject, ovViewportDashedLineScale, block);
}

Real64 VWViewportObj::GetHatchLineScale() const
{
	Real64 scale;
	TVariableBlock block(scale);
	gSDK->GetObjectVariable(fhObject, ovViewportHatchLineScale, block);
	block.GetReal64(scale);

	return scale;
}

void VWViewportObj::SetHatchLineScale(Real64 scale)
{
	TVariableBlock block(scale);
	gSDK->SetObjectVariable(fhObject, ovViewportHatchLineScale, block);
}

Real64 VWViewportObj::GetDesignTextScale() const
{
	Real64 scale;
	TVariableBlock block(scale);
	gSDK->GetObjectVariable(fhObject, ovViewportDesignTextScale, block);
	block.GetReal64(scale);

	return scale;
}

void VWViewportObj::SetDesignTextScale(Real64 scale)
{
	TVariableBlock block(scale);
	gSDK->SetObjectVariable(fhObject, ovViewportDesignTextScale, block);
}

Real64 VWViewportObj::GetSlashThicknessScale() const
{
	Real64 scale;
	TVariableBlock block(scale);
	gSDK->GetObjectVariable(fhObject, ovViewportSlashThicknessScale, block);
	block.GetReal64(scale);

	return scale;
}

void VWViewportObj::SetSlashThicknessScale(Real64 scale)
{
	TVariableBlock block(scale);
	gSDK->SetObjectVariable(fhObject, ovViewportSlashThicknessScale, block);
}

////////////////////////////////////////////////////////////////////////////////
// Override Manipulators
////////////////////////////////////////////////////////////////////////////////

bool VWViewportObj::GetUseDocumentClassVis() const
{
	Boolean useDocumentClassVis;
	TVariableBlock block(useDocumentClassVis);
	gSDK->GetObjectVariable(fhObject, ovViewportUseDocumentClassVis, block);
	block.GetBoolean(useDocumentClassVis);

	return useDocumentClassVis;
}

void VWViewportObj::SetUseDocumentClassVis(bool useDocumentClassVis)
{
	TVariableBlock block((Boolean) useDocumentClassVis);
	gSDK->SetObjectVariable(fhObject, ovViewportUseDocumentClassVis, block);
}

bool VWViewportObj::GetClassVisibility(InternalIndex classIndex, short& outVisibilityType) const
{
	return gSDK->GetViewportClassVisibility(fhObject, classIndex, outVisibilityType);
}

bool VWViewportObj::SetClassVisibility(InternalIndex classIndex, short inVisibilityType)
{
	return gSDK->SetViewportClassVisibility(fhObject, classIndex, inVisibilityType);
}

size_t VWViewportObj::GetNumClassOverrides() const
{
	return gSDK->GetNumViewportClassOverrides(fhObject);
}

InternalIndex VWViewportObj::GetClassOverrideClass(size_t index) const
{
	return gSDK->GetViewportClassOverrideClass(fhObject, index);
}

bool VWViewportObj::CreateClassOverride(InternalIndex classIndex)
{
	return gSDK->CreateViewportClassOverride(fhObject, classIndex);
}

bool VWViewportObj::GetClassOverride(InternalIndex classIndex, SViewportClassOverride& outOverride) const
{
	return gSDK->GetViewportClassOverride(fhObject, classIndex, outOverride);
}

bool VWViewportObj::SetClassOverride(InternalIndex classIndex, const SViewportClassOverride& inOverride)
{
	return gSDK->SetViewportClassOverride(fhObject, classIndex, inOverride);
}

bool VWViewportObj::RemoveClassOverride(InternalIndex classIndex)
{
	return gSDK->RemoveViewportClassOverride(fhObject, classIndex);
}

////////////////////////////////////////////////////////////////////////////////

size_t VWViewportObj::GetNumLightOverrides() const
{
	return gSDK->GetNumViewportLightOverrides(fhObject);
}

InternalIndex VWViewportObj::GetLightOverrideLight(size_t index) const
{
	return gSDK->GetViewportLightOverrideLight(fhObject, index);
}

bool VWViewportObj::CreateLightOverride(InternalIndex lightIndex)
{
	return gSDK->CreateViewportLightOverride(fhObject, lightIndex);
}

bool VWViewportObj::GetLightOverride(InternalIndex lightIndex, SViewportLightOverride& outOverride) const
{
	return gSDK->GetViewportLightOverride(fhObject, lightIndex, outOverride);
}

bool VWViewportObj::SetLightOverride(InternalIndex lightIndex, const SViewportLightOverride& inOverride)
{
	return gSDK->SetViewportLightOverride(fhObject, lightIndex, inOverride);
}

bool VWViewportObj::RemoveLightOverride(InternalIndex lightIndex)
{
	return gSDK->RemoveViewportLightOverride(fhObject, lightIndex);
}

////////////////////////////////////////////////////////////////////////////////

bool VWViewportObj::GetLayerVisibility(InternalIndex layerIndex, short& outVisibilityType) const
{
	MCObjectHandle layerHandle = gSDK->InternalIndexToHandle(layerIndex);
	return gSDK->GetViewportLayerVisibility(fhObject, layerHandle, outVisibilityType);
}

bool VWViewportObj::SetLayerVisibility(InternalIndex layerIndex, short inVisibilityType)
{
	MCObjectHandle layerHandle = gSDK->InternalIndexToHandle(layerIndex);
	return gSDK->SetViewportLayerVisibility(fhObject, layerHandle, inVisibilityType);
}

size_t VWViewportObj::GetNumLayerOverrides() const
{
	return gSDK->GetNumViewportLayerOverrides(fhObject);
}

InternalIndex VWViewportObj::GetLayerOverrideLayer(size_t index) const
{
	return gSDK->GetViewportLayerOverrideLayer(fhObject, index);
}

bool VWViewportObj::CreateLayerOverride(InternalIndex layerIndex)
{
	return gSDK->CreateViewportLayerOverride(fhObject, layerIndex);
}

bool VWViewportObj::GetLayerOverride(InternalIndex layerIndex, SViewportLayerOverride& outOverride) const
{
	return gSDK->GetViewportLayerOverride(fhObject, layerIndex, outOverride);
}

bool VWViewportObj::SetLayerOverride(InternalIndex layerIndex, const SViewportLayerOverride& inOverride)
{
	return gSDK->SetViewportLayerOverride(fhObject, layerIndex, inOverride);
}

bool VWViewportObj::RemoveLayerOverride(InternalIndex layerIndex)
{
	return gSDK->RemoveViewportLayerOverride(fhObject, layerIndex);
}

////////////////////////////////////////////////////////////////////////////////

size_t VWViewportObj::GetNumLayerPropertiesOverrides() const
{
	return gSDK->GetNumViewportLayerPropertiesOverrides(fhObject);
}

InternalIndex VWViewportObj::GetLayerPropertiesOverrideLayer(size_t index) const
{
	return gSDK->GetViewportLayerPropertiesOverrideLayer(fhObject, index);
}

bool VWViewportObj::CreateLayerPropertiesOverride(InternalIndex layerIndex)
{
	return gSDK->CreateViewportLayerPropertiesOverride(fhObject, layerIndex);
}

bool VWViewportObj::GetLayerPropertiesOverride(InternalIndex layerIndex, SViewportLayerPropertiesOverride& outOverride) const
{
	return gSDK->GetViewportLayerPropertiesOverride(fhObject, layerIndex, outOverride);
}

bool VWViewportObj::SetLayerPropertiesOverride(InternalIndex layerIndex, const SViewportLayerPropertiesOverride& inOverride)
{
	return gSDK->SetViewportLayerPropertiesOverride(fhObject, layerIndex, inOverride);
}

bool VWViewportObj::RemoveLayerPropertiesOverride(InternalIndex layerIndex)
{
	return gSDK->RemoveViewportLayerPropertiesOverride(fhObject, layerIndex);
}

////////////////////////////////////////////////////////////////////////////////

size_t VWViewportObj::GetNumLayerStackingOverrides() const
{
	return gSDK->GetNumViewportLayerStackingOverrides(fhObject);
}

InternalIndex VWViewportObj::GetLayerStackingOverrideLayer(size_t index) const
{
	return gSDK->GetViewportLayerStackingOverrideLayer(fhObject, index);
}

bool VWViewportObj::GetLayerStackingOverride(InternalIndex layerIndex, long& outStackingPosition) const
{
	return gSDK->GetViewportLayerStackingOverride(fhObject, layerIndex, outStackingPosition);
}

bool VWViewportObj::SetLayerStackingOverride(InternalIndex layerIndex, long inStackingPosition)
{
	return gSDK->SetViewportLayerStackingOverride(fhObject, layerIndex, inStackingPosition);
}

bool VWViewportObj::RemoveLayerStackingOverride(InternalIndex layerIndex)
{
	return gSDK->RemoveViewportLayerStackingOverride(fhObject, layerIndex);
}

////////////////////////////////////////////////////////////////////////////////
// Handle Inspectors
////////////////////////////////////////////////////////////////////////////////

/* Returns true if the provided handle points to a viewport
 * Returns false otherwise
 */
bool VWViewportObj::IsViewportObject(MCObjectHandle hObj)
{
	return gSDK->GetObjectTypeN(hObj) == kViewportNode;
}

/* Returns true if the provided handle points to a design layer viewport
 * Returns false otherwise
 */
bool VWViewportObj::IsDesignLayerViewportObject(MCObjectHandle hObj)
{
	return gSDK->IsDesignLayerViewport(hObj);
}