//
//	Copyright  Nemetschek Vectorworks, Inc.
//	Use of this file is governed by the Nemetschek Vectorworks SDK License Agreement
//	http://developer.vectorworks.net/index.php?title=Vectorworks_SDK_License
//

#include "StdHeaders.h"
#include "OverriddenControl.h"
#include "MiniCadCallBacks.h"
#if _MINICAD_
#include "DrawPadFactory.h"
#endif

#include <map>
std::map<void *, OverriddenLayoutControl *> controlMap;
std::map<void *, OverriddenLayoutControl *> :: const_iterator controlIterator;
OverriddenLayoutControl * overriddenCtrlUnderMouse = nil; // which control is under the mouse

OverriddenLayoutControl::OverriddenLayoutControl()
{
#if _MINICAD_
	fpDrawPad = nil;
#endif
	fDlogID = 0;
	fControlID = 0;
	fInitialized = false;
	fUsesDrawPad = false;
	fDrawPadInitialized = false;
	fIdentifier = nil;
	mouseEntered = false;
	fDrawRectInitialized = false;
	fDrawRect.SetLTRB(0, 0, 0, 0);
	
#if !GS_WIN
	initialMouseMove = false;
#endif
}

#if GS_WIN
LRESULT CALLBACK OverriddenLayoutControlProc(HWND controlHWND, UINT message, WPARAM wParam, LPARAM lParam)
{
	controlIterator = controlMap.find((void *)controlHWND);
	if (controlIterator == controlMap.end())
		return false;
	
	OverriddenLayoutControl *control = controlIterator->second;
	
	switch (message) {
		case WM_PAINT: {
			PAINTSTRUCT ps;
			BeginPaint(controlHWND, &ps);
			control->SetupDraw();
			
			control->OnDraw();
			control->FinishDraw();
			EndPaint(controlHWND, &ps);
			return true;
		}
		case WM_LBUTTONDOWN: {
			SetCapture(controlHWND);
			ViewPt p;
			p.x = LOWORD(lParam); 
			p.y = HIWORD(lParam);
			control->OnLButtonDown(p.x, p.y);
			return true;
		}
		case WM_LBUTTONUP: {
			ReleaseCapture();
			ViewPt p;
			p.x = LOWORD(lParam); 
			p.y = HIWORD(lParam);
			control->OnLButtonUp(p.x, p.y);
			return true;
		}
		case WM_MBUTTONDOWN: {
			SetCapture(controlHWND);
			ViewPt p;
			p.x = LOWORD(lParam); 
			p.y = HIWORD(lParam);
			control->OnMButtonDown(p.x, p.y);
			return true;
		}
		case WM_MBUTTONUP: {
			ReleaseCapture();
			ViewPt p;
			p.x = LOWORD(lParam); 
			p.y = HIWORD(lParam);
			control->OnMButtonUp(p.x, p.y);
			return true;
		}
			
		case WM_RBUTTONDOWN: {
			SetCapture(controlHWND);
			ViewPt p;
			p.x = LOWORD(lParam); 
			p.y = HIWORD(lParam);
			control->OnRButtonDown(p.x, p.y);
			return true;
		}
		case WM_RBUTTONUP: {
			ReleaseCapture();
			ViewPt p;
			p.x = LOWORD(lParam); 
			p.y = HIWORD(lParam);
			control->OnRButtonUp(p.x, p.y);
			return true;
		}
		case WM_MOUSEWHEEL: {
			short zDelta = GET_WHEEL_DELTA_WPARAM(wParam);
			POINT pt;
			pt.x = LOWORD(lParam); 
			pt.y = HIWORD(lParam); 
			ScreenToClient(controlHWND, &pt);
			control->OnMouseWheel(pt.x, pt.y, (int)(zDelta / WHEEL_DELTA));
			return true;
		}
			
		case WM_MOUSELEAVE: {
			if( control->mouseEntered ) {
				control->OnMouseLeave();
			}
			control->mouseEntered = false;
			overriddenCtrlUnderMouse = nil;
			
			return true;
		}
		case WM_MOUSEMOVE: {
			if( !control->mouseEntered ) {
				overriddenCtrlUnderMouse = control;
				control->mouseEntered = true;
				TRACKMOUSEEVENT trackEvent;
				trackEvent.cbSize = sizeof(TRACKMOUSEEVENT);
				trackEvent.dwFlags = TME_LEAVE;
				trackEvent.hwndTrack = controlHWND;
				trackEvent.dwHoverTime = HOVER_DEFAULT;
				TrackMouseEvent(&trackEvent);
				control->OnMouseEnter();
			}
			ViewPt p;
			p.x = LOWORD(lParam); 
			p.y = HIWORD(lParam);
			control->OnMouseMove(p.x, p.y);
			return true; 
		}
	}
	
	return DefWindowProc(controlHWND, message, wParam, lParam);
}


void OverriddenLayoutControl::Attach(int dlogID, int controlID, bool wantDrawPad)
{	
	HWND hwDlg = (HWND) GS_OverrideControl(gCBP, dlogID, controlID, OverriddenLayoutControlProc);
	fIdentifier = (void *) GetDlgItem(hwDlg, controlID);
	fDlogID = dlogID;
	fControlID = controlID;
	mouseEntered = false;
	overriddenCtrlUnderMouse = nil;
	
	controlMap[fIdentifier] = this;
	if (wantDrawPad) {
		fUsesDrawPad = true;
		fDrawPadInitialized = false;
	}
	else {
		fUsesDrawPad = false;
	}
}

void OverriddenLayoutControl::SetupDraw()
{
	// The following SetDrawRect call is very important, otherwise fDrawRect will not be initialized 
	// for the OnInit call - resulting in no image in the Edit Texture dialog for example. <Biplab 05/03/07>
	RECT r;
	GetClientRect((HWND) fIdentifier, &r);
	ViewRect drawRect((ViewCoord)r.left, (ViewCoord)r.top, (ViewCoord)r.right, (ViewCoord)r.bottom);
	SetDrawRect(drawRect);
	
	if (!fInitialized) {
		
		if (fUsesDrawPad) {
			if (!fDrawPadInitialized) {
#if _MINICAD_
				fpDrawPad = CreateDrawPad(TWindowDev((unsigned long) fIdentifier));
#endif
				fDrawPadInitialized = true;
			}
			
#if _MINICAD_		
			if (fpDrawPad) {
				
				RECT r;
				ViewRect vr;
				
				GetClientRect((HWND) fIdentifier, &r);
				vr.SetLTRB((ViewCoord) r.left,  (ViewCoord) r.top, (ViewCoord) r.right, (ViewCoord) r.bottom);
				fpDrawPad->SetClipRect(vr);
				fpDrawPad->SetPen(1, kGSColorBlack, kGSColorWhite, kSolidForePattern);
			}
#endif
		}
		
		fInitialized = true;
		
		this->OnInit();
	}
	
	
#if _MINICAD_	
	if (fUsesDrawPad)
		fpDrawPad->CaptureDevice();
#endif
}

void OverriddenLayoutControl::FinishDraw()
{
#if _MINICAD_	
	if (fUsesDrawPad)
		fpDrawPad->ReleaseDevice();	
#endif
}

#else	// !GS_WIN

typedef struct {
	Point		wherePort;
	Point		whereLocal;
	EventRecord	macEvent;
	Boolean		delaySelect;
} SMouseEvent;

bool UsingCocoaForDialog(int dlogID)
{
	return gSDK->Kludge(650, &dlogID, nil);
}

struct CocoaCustomControlDrawParam
{
	ViewRect viewRect;
	CGContextRef contextRef;
};

void OverriddenLayoutControlProc(long message, void *ioParam, void *pane, int dialogID, int componentID) {
	const signed long msg_DrawOrPrint	= 811;	// Rect* (frame of Pane)
	const signed long msg_Click			= 812;	// SMouseEvent*
	const signed long msg_AdjustCursor	= 813;	// SMouseEvent*
	const signed long msg_MouseUp       = 850;	// SMouseEvent*
	
	componentID;
	dialogID;
	
	controlIterator = controlMap.find(pane);
	if (controlIterator == controlMap.end())
		return;
	
	OverriddenLayoutControl *control = controlIterator->second;
	
	switch (message) {
		case msg_DrawOrPrint: {
			
			if (UsingCocoaForDialog(dialogID))
				control->fCocoaGraphicsContext = ((CocoaCustomControlDrawParam*) ioParam)->contextRef;
			
			ViewRect* macDrawRect = (ViewRect*) ioParam;
			ViewRect drawRect(macDrawRect->left, macDrawRect->top, macDrawRect->right, macDrawRect->bottom);
			control->SetDrawRect(drawRect);
			
			control->SetupDraw();
			control->OnDraw();
			control->FinishDraw();
			break;
		}
		case msg_Click: {
			control->OnLButtonDown(
								   ((SMouseEvent*)ioParam)->whereLocal.h - control->fDrawRect.left,  
								   ((SMouseEvent*)ioParam)->whereLocal.v - control->fDrawRect.top);
			break;
		}
		case msg_AdjustCursor: {
			if (control->fInitialized) {
				
				control->OnMouseMove(
									 ((SMouseEvent*)ioParam)->whereLocal.h - control->fDrawRect.left,  
									 ((SMouseEvent*)ioParam)->whereLocal.v - control->fDrawRect.top);
			} else {
				control->initialMouseMove = true;
				control->initialMouseMoveX = ((SMouseEvent*)ioParam)->whereLocal.h - control->fDrawRect.left;
				control->initialMouseMoveY = ((SMouseEvent*)ioParam)->whereLocal.v - control->fDrawRect.top;
			}
			
			break;
		}
		case msg_MouseUp: {
			control->OnLButtonUp(
								 ((SMouseEvent*)ioParam)->whereLocal.h - control->fDrawRect.left,  
								 ((SMouseEvent*)ioParam)->whereLocal.v - control->fDrawRect.top);
			break;
		}
			
		case 100: {
			SMouseEvent* mouseEvent = (SMouseEvent*) ioParam;
			short actualX = mouseEvent->wherePort.h - control->fDrawRect.left;
			
			// We have to flip the v coordinate since for mouse wheel the "local" coordinates are used for the amount of scroll...
			short controlHeight = control->fDrawRect.bottom - control->fDrawRect.top;
			short actualY = (controlHeight - mouseEvent->wherePort.v) - control->fDrawRect.top;
			
			control->OnMouseWheel(actualX, actualY, mouseEvent->whereLocal.v);
			break;
		}
		case 101:
			control->OnMouseEnter();
			break;
		case 102:
			control->OnMouseLeave();
			break;
	}
	
	return;
}

void* GetPane(int dlogID, int ctrlID);

void OverriddenLayoutControl::Attach(int dlogID, int controlID, bool wantDrawPad)
{
    GS_OverrideControl(gCBP, dlogID, controlID, (void *)OverriddenLayoutControlProc);
	fIdentifier = GetPane(dlogID, controlID);
	fDlogID = dlogID;
	fControlID = controlID;
	
	
	controlMap[fIdentifier] = this;
	if (wantDrawPad) {
		fUsesDrawPad = true;
		fDrawPadInitialized = false;
	}
	else {
		fUsesDrawPad = false;
	}
	
	/*HIViewRef view = HIViewGetRoot(window);
	 HIViewTrackingAreaRef trackingArea;
	 HIViewNewTrackingArea(view, NULL, (HIViewTrackingAreaID)this, &trackingArea);
	 */
	
	if (UsingCocoaForDialog(dlogID))	
        return;								//  (no panes, no carbon event handlers, etc.)
    DSTOP((kTArdoin, "OverriddenLayoutControl::Attach() - Non Cocoa dialog passed in."));
}

void OverriddenLayoutControl::SetupDraw()
{
	if (!fInitialized) {
		
		if (fUsesDrawPad) {
			
			if (UsingCocoaForDialog(fDlogID)) {
				// draw pad gets set up for each draw in Cocoa dialogs, not just initial draw
			}
			
			else {
                DSTOP((kTArdoin, "OverriddenLayoutControl::SetupDraw() - Non Cocoa dialog passed in."));
            }
		}
		
		fInitialized = true;
		
		this->OnInit();
		
		if( initialMouseMove ) {
			overriddenCtrlUnderMouse = this;
			mouseEntered = true;
			OnMouseEnter();
			
			OnMouseMove(initialMouseMoveX, initialMouseMoveY);
			initialMouseMove = false;
		}
	}
	
#if _MINICAD_
	if (fUsesDrawPad) { 
		if (UsingCocoaForDialog(fDlogID)) {
			fpDrawPad = new TDrawPadQuartz2D(fCocoaGraphicsContext, fDrawRect);
		}
		else {
            DSTOP((kTArdoin, "OverriddenLayoutControl::SetupDraw() - Non Cocoa dialog used."));
		}
	}
#endif
}

void OverriddenLayoutControl::FinishDraw()
{
#if _MINICAD_
	if (fUsesDrawPad) {
		if (UsingCocoaForDialog(fDlogID)) {
			delete fpDrawPad;
			fpDrawPad = NULL;
		}
		else {
            DSTOP((kTArdoin, "OverriddenLayoutControl::FinishDraw() - Non Cocoa dialog used."));
		}
	}
#endif
}

#endif	// !GS_WIN


void OverriddenLayoutControl::Detach()
{
	controlMap.erase(fIdentifier);
#if _MINICAD_
	if (fUsesDrawPad && fpDrawPad != NULL)
		delete fpDrawPad;
	fpDrawPad = NULL;
#endif
	fInitialized = false;
	OnDetatch();
}
