//
//	Copyright  Nemetschek Vectorworks, Inc.
//	Use of this file is governed by the Nemetschek Vectorworks SDK License Agreement
//	http://developer.vectorworks.net/index.php?title=Vectorworks_SDK_License
//

#include "StdHeaders.h"
#include <vector>
#include "VWFC/PluginSupport/VWExtensions.h"
#include "VWFC/PluginSupport/VWExtensionParametric.h"

using namespace VWFC::PluginSupport;

// ------------------------------------------------------------------------------------------------------
bool SSTRResource::operator<(const SSTRResource& other) const
{
	if ( fListID == other.fListID ) {
		return fIndex < other.fIndex;
	}
	else {
		return fListID < other.fListID;
	}
}

// ------------------------------------------------------------------------------------------------------
bool STEXTResource::operator<(const STEXTResource& other) const
{
	return fID < other.fID;
}

// ------------------------------------------------------------------------------------------------------
VWExtensionParametric::VWExtensionParametric(CallBackPtr cbp,
								const SParametricDef& parametricDef,
								const SParametricParamDef* arrParamDefs,
								const SParametricParamChc* arrParamChcs)
	: fCBP( cbp )
	, fParametricDef( parametricDef )
	, farrParamDefs( arrParamDefs )
	, farrParamChcs( arrParamChcs )
	, fParametersSink( NULL )
	, fLegacyEventSink( NULL )
	, fProtectionObtained( false )
	, fHasProtection( false )
	, fProtectionSink( NULL )
{
	gCBP				= fCBP;
	fParamsCount		= 0;
	
	fstrLocalizedName.Load( fParametricDef.fLocalizedName.fListID, fParametricDef.fLocalizedName.fIndex );

	// first count the number of things we will need to store
	const SParametricParamDef*	currParam	= farrParamDefs;
	while ( currParam && currParam->fUniversalName ) 
	{
		fParamsCount++;
		currParam++;
	}

	// These arrays will be used to store temporary ID/Index pairs.
	const int numStaticElements = 256;
	short idsStaticArray[numStaticElements], *idsArray = idsStaticArray;
	short indexStaticArray[numStaticElements], *indexArray = indexStaticArray;
	SSTRResource localizedNameIDIndexPairsStaticArray[numStaticElements], *localizedNameIDIndexPairs = localizedNameIDIndexPairsStaticArray;
	if (fParamsCount > numStaticElements)
	{	// most of the time, static storage is used. If param count is too big to fit, we use heap storage.
		idsArray = new short[fParamsCount];
		indexArray = new short[fParamsCount];
		localizedNameIDIndexPairs = new SSTRResource[fParamsCount];
	}

	int numItemsNotCached = 0;
	for (int curParamIndex = 0; curParamIndex < (int) fParamsCount; ++curParamIndex)
	{
		currParam = &farrParamDefs[curParamIndex];
		TResStringCache::iterator	resIt	= fResStringChache.find( currParam->fLocalizedName );
		if ( resIt == fResStringChache.end() ) 
		{	// if not found in cache already
			localizedNameIDIndexPairs[numItemsNotCached] = currParam->fLocalizedName;
			idsArray[numItemsNotCached] = currParam->fLocalizedName.fListID;
			indexArray[numItemsNotCached] = currParam->fLocalizedName.fIndex;
			++numItemsNotCached;
		}
	}

	// Get all these strings from the resource at once...
	TXStringArray strings(numItemsNotCached);
	gSDK->GSGetResourceStringArray(idsArray, indexArray, numItemsNotCached, strings, true, eDontWarnAboutEmptyStringResult);

	// Insert them in the cache.
	for (int i = 0; i < numItemsNotCached; ++i)
	{
		fResStringChache.insert( TResStringCache::value_type(localizedNameIDIndexPairs[i], strings[i]) );
	}

	// ------------------------------------- 
	// Now do the same for the chc array, which to my knowledge isn't necessarily the same as fParamsCount - if it is, some housekeeping can be removed below
	// First, count them.
	int numChcs = 0;
	const SParametricParamChc*	currChc		= farrParamChcs;
	while( currChc && currChc->fChcID ) 
	{
		++numChcs;
		currChc++;
	}

	if (numChcs > (int) fParamsCount)
	{	// we may need to reallocate our arrays
		if (numChcs > numStaticElements)
		{	
			// deallocate if necessary
			if (idsArray != idsStaticArray)
			{
				delete [] idsArray;
				delete [] indexArray;
				delete [] localizedNameIDIndexPairs;
			}
			idsArray = new short[numChcs];
			indexArray = new short[numChcs];
			localizedNameIDIndexPairs = new SSTRResource[numChcs];
		}
	}

	strings.Clear();
	numItemsNotCached = 0;
	for (int curChcIndex = 0; curChcIndex < numChcs; ++curChcIndex)
	{
		const SParametricParamChc* currChc = &farrParamChcs[curChcIndex];
		TResStringCache::iterator	resIt	= fResStringChache.find( currChc->fLocalizedChoice );
		if ( resIt == fResStringChache.end() ) {	// not already cached
			localizedNameIDIndexPairs[numItemsNotCached] =  currChc->fLocalizedChoice;
			idsArray[numItemsNotCached] = currChc->fLocalizedChoice.fListID;
			indexArray[numItemsNotCached] = currChc->fLocalizedChoice.fIndex;
			++numItemsNotCached;
		}

		TChcCache::const_iterator	chcIt	= fParamChcCache.find( currChc->fChcID );	
		if ( chcIt == fParamChcCache.end() ) {
			// all concequitive chc values have to have the same ID
			fParamChcCache.insert( TChcCache::value_type( currChc->fChcID, currChc ) );
		}

		currChc++;
	}

	// Get all these strings from the resource at once...
	strings.SetSize(numItemsNotCached);
	gSDK->GSGetResourceStringArray(idsArray, indexArray, numItemsNotCached, strings, true, eDontWarnAboutEmptyStringResult);

	// Insert them in the cache.
	for (int i = 0; i < numItemsNotCached; ++i)
	{
		fResStringChache.insert( TResStringCache::value_type( localizedNameIDIndexPairs[i], strings[i]) );
	}

	// deallocate if necessary
	if (idsArray != idsStaticArray)
	{
		delete [] idsArray;
		delete [] indexArray;
		delete [] localizedNameIDIndexPairs;
	}

}

VWExtensionParametric::~VWExtensionParametric()
{
}

void VWExtensionParametric::SetLocalizedParamDef( const short listID )
{
	const SParametricParamDef*	currParam	= farrParamDefs;
	while ( currParam && currParam->fUniversalName ) {
		TXString	str;
		gSDK->GSGetResourceString( listID, currParam->fLocalizedName.fIndex, str, true, eDontWarnAboutEmptyStringResult );

		if ( str.IsEmpty() == false ) {
			TResStringCache::iterator	resIt	= fResStringChache.find( currParam->fLocalizedName );
			if ( resIt != fResStringChache.end() ) {
				resIt->second = str;
			}
		}
		currParam++;
	}
}

void VWExtensionParametric::SetLocalizedParamChc( const short listID )
{
	const SParametricParamChc*	currChc		= farrParamChcs;
	while( currChc && currChc->fChcID ) {
		TXString	str;
		gSDK->GSGetResourceString( listID, currChc->fLocalizedChoice.fIndex, str, true, eDontWarnAboutEmptyStringResult );

		if ( str.IsEmpty() == false ) {
			TResStringCache::iterator	resIt	= fResStringChache.find( currChc->fLocalizedChoice );
			if ( resIt != fResStringChache.end() ) {
				resIt->second = str;
			}
		}
		currChc++;
	}

}

void VWExtensionParametric::OnRefCountZero()
{
	if ( fLegacyEventSink ) {
		delete fLegacyEventSink;
		fLegacyEventSink	= NULL;
	}

	if ( fParametersSink ) {
		delete fParametersSink;
		fParametersSink	= NULL;
	}

	if ( fProtectionSink ) {
		delete fProtectionSink;
		fProtectionSink	= NULL;
	}

	for(TSinksMap::iterator it=fmapSinks.begin(); it!=fmapSinks.end(); it++) {
		IEventSink*	theSink	= it->second;
		if ( theSink ) {
			delete theSink;
		}
	}

	fmapSinks.clear();

	// unload resources
	::GS_VWUnloadInterfaceResource( this );
}

IEventSink* VWExtensionParametric::QueryEventSink(const TSinkIID& iid)
{
	if ( fProtectionObtained == false ) {
		fHasProtection			= this->GetProtectionSinkIID( fProtectionSinkIID );
		fProtectionObtained		= true;
	}

	IEventSink*	pResult	= NULL;
	if ( IID_ParametricParamsProvider == iid ) {
		if ( fParametersSink == NULL ) {
			fParametersSink	= new CParametersProvider( this );
		}

		if ( fParametersSink ) {
			fParametersSink->AddRef();
		}

		pResult	= fParametersSink;
	}
	else if ( IID_LegacyEventSink == iid ) {
		if ( fLegacyEventSink == NULL ) {
			fLegacyEventSink	= this->CreateLegacyEventSink( this );
			if ( fLegacyEventSink ) {
				fLegacyEventSink->AddRef();
				VWParametric_EventSink*	vwfcSink	= dynamic_cast<VWParametric_EventSink*>( fLegacyEventSink );
				if ( vwfcSink ) {
					vwfcSink->SetUniversalName( this->GetUniversalName() );
				}
			}
		}
		pResult	= fLegacyEventSink;
	}
	else if ( fHasProtection && fProtectionSinkIID == iid ) {
		if ( fProtectionSink == NULL ) {
			fProtectionSink	= this->CreateProtectionSink( this );
			if ( fProtectionSink == NULL ) {
				fHasProtection	= false;
			}
		}

		if ( fProtectionSink ) {
			fProtectionSink->AddRef();
			pResult	= fProtectionSink;
		}
	}
	else {
		IEventSink*			theSink	= NULL;
		TSinksMap::iterator	it		= fmapSinks.find( iid );
		if ( it != fmapSinks.end() ) {
			theSink	= it->second;
		}
		else {
			fDefinedSink		= NULL;
			fDefinedSinkIID		= iid;
			this->DefineSinks();
			if ( fDefinedSink ) {
				fmapSinks.insert( TSinksMap::value_type( iid, fDefinedSink ) );
				theSink			= fDefinedSink;
				fDefinedSink	= NULL;
			}
		}

		if ( theSink ) {
			theSink->AddRef();
			pResult			= theSink;
		}
	}

	return pResult;
}

void VWExtensionParametric::DefineSinks()
{
}

bool VWExtensionParametric::GetProtectionSinkIID(TSinkIID& outIID)
{
	return false;
}

IProviderProtection* VWExtensionParametric::CreateProtectionSink(IVWUnknown* parent)
{
	return NULL;
}

const TXString& VWExtensionParametric::GetLocalizedName()
{
	return fstrLocalizedName;
}

EParametricSubType VWExtensionParametric::GetObjectType()
{
	return fParametricDef.fSubType;
}

void VWExtensionParametric::GetResetModeOnMove(bool& outOnMove, bool& outOnRotate)
{
	outOnMove		= fParametricDef.fResetOnMove;
	outOnRotate		= fParametricDef.fResetOnRotate;
}
void VWExtensionParametric::GetWallInsertMode(bool& outOnEdge, bool& outNoBreak, bool& outHalfBreak, bool& outHideCaps)
{
	outOnEdge		= fParametricDef.fWallInsertOnEdge;
	outNoBreak		= fParametricDef.fWallInsertNoBreak;
	outHalfBreak	= fParametricDef.fWallInsertHalfBreak;
	outHideCaps		= fParametricDef.fWallInsertHideCaps;
}

// --------------------------------------------------------------------------------------------------------
VWExtensionParametric::CParametersProvider::CParametersProvider(VWExtensionParametric* parent)
	: fParent( parent )
{
}

VWExtensionParametric::CParametersProvider::~CParametersProvider()
{
}

size_t VWExtensionParametric::CParametersProvider::GetParamsCount()
{
	return fParent->fParamsCount;
}

void VWExtensionParametric::CParametersProvider::GetParamAt(size_t paramIndex, TXString& outUniversalName, EFieldStyle& outType)
{
	ASSERTN( kVStanev, paramIndex < fParent->fParamsCount );
	const SParametricParamDef&	thisParam	= fParent->farrParamDefs[ paramIndex ];

	outUniversalName	= thisParam.fUniversalName;
	outType				= thisParam.fFieldStyle;

	outUniversalName.AnsiToMac();
}

const TXString& VWExtensionParametric::CParametersProvider::GetParamNameAt(size_t paramIndex)
{
	ASSERTN( kVStanev, paramIndex < fParent->fParamsCount );
	const SParametricParamDef&	thisParam	= fParent->farrParamDefs[ paramIndex ];

	TResStringCache::const_iterator	resIt	= fParent->fResStringChache.find( thisParam.fLocalizedName );
	return (VERIFYN( kVStanev, resIt != fParent->fResStringChache.end() )) ? resIt->second : fParent->fEmptyString;
}

void VWExtensionParametric::CParametersProvider::GetParamDefValueImperial(size_t paramIndex, TXString& outValue)
{
	ASSERTN( kVStanev, paramIndex < fParent->fParamsCount );
	const SParametricParamDef&	thisParam	= fParent->farrParamDefs[ paramIndex ];

	this->GetDefaultValue( thisParam.fDefaultImperial, outValue );
}

void VWExtensionParametric::CParametersProvider::GetParamDefValueMetric(size_t paramIndex, TXString& outValue)
{
	ASSERTN( kVStanev, paramIndex < fParent->fParamsCount );
	const SParametricParamDef&	thisParam	= fParent->farrParamDefs[ paramIndex ];

	this->GetDefaultValue( thisParam.fDefaultMetric, outValue );
}

void VWExtensionParametric::CParametersProvider::GetParamChoices(size_t paramIndex, TXStringArray& outArrUniversalNames, TXStringArray& outArrDisplayNames)
{
	outArrUniversalNames.Clear();
	outArrDisplayNames.Clear();

	ASSERTN( kVStanev, paramIndex < fParent->fParamsCount );
	const SParametricParamDef&	thisParam	= fParent->farrParamDefs[ paramIndex ];

	if ( thisParam.fChoicesChcID > 0 ) {
		TChcCache::const_iterator	chcIt	= fParent->fParamChcCache.find( thisParam.fChoicesChcID );
		if ( chcIt != fParent->fParamChcCache.end() ) {
			const SParametricParamChc*	currChc	= chcIt->second;
			while( currChc && thisParam.fChoicesChcID == currChc->fChcID && currChc->fUniversalChoice ) {

				TXString	universalChoice	= currChc->fUniversalChoice;
				universalChoice.AnsiToMac();
				outArrUniversalNames.Append( universalChoice );

				TResStringCache::const_iterator	resIt	= fParent->fResStringChache.find( currChc->fLocalizedChoice );
				outArrDisplayNames.Append( (VERIFYN( kVStanev, resIt != fParent->fResStringChache.end() )) ? resIt->second : "" );

				currChc++;
			}
		}
	}
}

void VWExtensionParametric::CParametersProvider::GetDefaultValue(const char* szDefaultValue, TXString& outValue)
{
	if ( szDefaultValue[0] == 0x01 ) {
		if ( VERIFYN( kVStanev, szDefaultValue[1] != 0x00 ) ) {
			ASSERTN( kVStanev, szDefaultValue[1] == '@' );
			TXString	resValue	= szDefaultValue + 2;
			short		resListID	= resValue.atoi();
			int			indexAt		= resValue.Find( ',' );
			if ( indexAt >= 0 ) {
				resValue	= resValue.Mid( indexAt + 1 );
				short	resIndex	= resValue.atoi();

				outValue.Load( resListID, resIndex, eAllowEmptyResult );
			}
		}
	}
	else {
		outValue	= szDefaultValue;
	}
}


// --------------------------------------------------------------------------------------------------------
// predefine these hidden memory managment functions
namespace VWFC
{
	void	AllocateMemoryManagers(VWFC::MemoryManager::MManager** ppOutOldManager, VWFC::MemoryManager::MManager** ppOutOldVWManager);
	void	FreeMemoryManagers(VWFC::MemoryManager::MManager* pOldManager, VWFC::MemoryManager::MManager* pOldVWManager);
}

// ------------------------------------------------------------------------------------------------------
VWParametric_EventSink::VWParametric_EventSink(IVWUnknown* parent)
	: VCOMImpl<ILegacyEventSink>( parent )
{
	fhObject			= NULL;
	fbInitialized		= false;
	fObjectEditReason	= kObjectEdit_ExplicitEdit;
}

VWParametric_EventSink::~VWParametric_EventSink()
{
	this->DeinitializeExt();
}

void VWParametric_EventSink::InitializeExt()
{
}

void VWParametric_EventSink::DeinitializeExt()
{
}

void VWParametric_EventSink::GetUserData(void) const
{
	// DO NOT USE THIS
	// This is to prevent old code from using it
}

void VWParametric_EventSink::SetUserData(void) const
{
	// DO NOT USE THIS
	// This is to prevent old code from using it
}

void VWParametric_EventSink::SetUniversalName(const TXString& name)
{
	fUnivName	= name;
}

long VWParametric_EventSink::LegacyMain(long action, long object, long message)
{
	long	reply	= kParametricNoErr;

	// create the memory namagers
	// and verify their presents
	VWFC::MemoryManager::MManager* pOldManager		= NULL;
	VWFC::MemoryManager::MManager* pOldVWManager	= NULL;
	VWFC::AllocateMemoryManagers( & pOldManager, & pOldVWManager );

	MCObjectHandle	saved_fhObject	= fhObject;

	VWFC_VERIFY( VWFC::GetMManager() != NULL && VWFC::GetVWMManager() != NULL );
	if ( VWFC::GetMManager() != NULL && VWFC::GetVWMManager() != NULL ) {

		fhObject		= (MCObjectHandle) object;

		if ( ! fbInitialized ) {
			this->InitializeExt();
			fbInitialized	= true;
		}

		switch (action) {
			case kParametricRecalculate:
				reply	= (long) this->Recalculate();
				break;
			case kParametricPreference:
				reply	= (long) this->Preference();
				break;
			case kObjOnInitXProperties:
				reply	= (long) this->OnInitXProperties( (CodeRefID) message );
				break;
			case kObjOnSpecialEditID:
				fObjectEditReason	= (TObjectEdiReason) message;
				reply	= (long) this->OnSpecialEditID( );
				break;
			case kObjOnObjectUIButtonHit:
				reply	= (long) this->OnObjectUIButtonHit( message );
				break;
			case kObjOnWidgetPrep: {
					OnObjectWidgetPrepCall* pThisCall		= (OnObjectWidgetPrepCall*) message;
					IWidgetsProvider*       pWidgetProvider = pThisCall ? (IWidgetsProvider*) pThisCall->fpWidgetProvider : NULL;
					if ( pWidgetProvider && this->OnWidgetUpdate( pWidgetProvider ) ) {
						reply	= kObjectEventHandled;
					}
				} break;
			case kObjOnAddState: {
					ObjectState*	pObjState	= (ObjectState*) message;
					if ( pObjState ) {
						reply	= (long) this->OnAddState( *pObjState );
					}
				} break;
			case kObjOnCursor_Draw:
			case kObjOnCursor_MouseDown:
			case kObjOnCursor_Complete:
			case kObjOnCursor_Cancel: 
			case kObjOnCursor_MouseMove:
			case kObjOnCursor_CustomBarEvent:
				reply = this->OnCursorEvent(action, message);
				break;
			case kObjOnGetToolName: {
					TXString*	pToolName	= (TXString*) message;
					if ( pToolName ) {
						*pToolName	= this->OnGetToolName();
						if ( ! pToolName->IsEmpty() ) {
							reply		= kObjectEventHandled;
						}
					}
				} break;
			case kObjOnContextMenuInit: {
					const ObjectContextMenuEvent* pEventData           = (const ObjectContextMenuEvent*) message;
					IContextMenuProvider*         pContextMenuProvider = (pEventData && pEventData->fData)
																   ? (IContextMenuProvider*) pEventData->fData : NULL;
					if ( pContextMenuProvider ) {
						this->OnContextMenuInit( pContextMenuProvider );
					}
				} break;
			case kObjOnContextMenuEvent: {
					const ObjectContextMenuEvent* pEventData    = (const ObjectContextMenuEvent*) message;
					if ( pEventData ) {
						this->OnContextMenuEvent( *pEventData );
					}
				} break;
			case kObjOnEyedropperPrepareCopy: {
					this->OnEyedropperPrepareCopy( (MCObjectHandle) message );
				} break;
			case kObjOnEyedropperAfterCopy: {
					this->OnEyedropperAfterCopy( (MCObjectHandle) message );
				} break;
			case kObjOnGetSpecificGeometry: {
					ObjectGetSpecificGeometryCall*	data	= (ObjectGetSpecificGeometryCall*) message;
					if ( data ) {
						data->fhGeometry	= (long) this->OnGetSpecificGeometry( (ObjectGetSpecificGeometryCall::ESpecifier) data->fSpecifier );
					}
				} break;

			default:
				reply = this->OnDefaultEvent( action, message );
				break;
		}

		// dump local memory leaks
		if ( VWFC::GetMManager() ) {
			VWFC::GetMManager()->DumpMemory( fUnivName );
		}
	}

	fhObject	= saved_fhObject;

	// release the memory managers
	VWFC::FreeMemoryManagers( pOldManager, pOldVWManager );

	return reply;
}

// The Plug-in Object code should re-create all its component objects using current parameter values.
EObjectEvent VWParametric_EventSink::Recalculate()
{
	return kObjectEventNoErr;
}

// Object Definition Procedure (ODP) may optionally present its own preferences dialog, or let VW pose the default dialog.
EObjectEvent VWParametric_EventSink::Preference()
{
	return kObjectEventNotImplemented;
}

// The Plug-in Object may add extended properties.
EObjectEvent VWParametric_EventSink::OnInitXProperties(CodeRefID /*objectID*/)
{
	return kObjectEventNoErr;
}

// The Plug-in Object called with kParametricOnMove when kHasMoveDependancy extended property set
EObjectEvent VWParametric_EventSink::OnSpecialEditID()
{
	return kObjectEventNoErr;
}

// The user has pressed a button in the Properties UI of an object
EObjectEvent VWParametric_EventSink::OnObjectUIButtonHit(long /*buttonID*/)
{
	return kObjectEventNoErr;
}

bool VWParametric_EventSink::OnWidgetUpdate(IWidgetsProvider* pWidgetProvider)
{
	return false;
}

EObjectEvent VWParametric_EventSink::OnAddState(ObjectState& stateInfo)
{
	return kObjectEventNoErr;
}

long VWParametric_EventSink::OnCursorEvent(ObjectEventID action, long message)
{
    return(kObjectEventNotHandled);
}

long VWParametric_EventSink::OnDefaultEvent(long action, long message)
{
	return 0;
}

TXString VWParametric_EventSink::OnGetToolName() const
{
	return TXString();
}

void VWParametric_EventSink::OnContextMenuInit(IContextMenuProvider* pContextMenuProvider)
{
}

void VWParametric_EventSink::OnContextMenuEvent(const ObjectContextMenuEvent& eventData)
{
}

void VWParametric_EventSink::OnEyedropperPrepareCopy(MCObjectHandle hObjRecordCopy)
{
}

void VWParametric_EventSink::OnEyedropperAfterCopy(MCObjectHandle hObjRecordCopy)
{
}

MCObjectHandle VWParametric_EventSink::OnGetSpecificGeometry(ObjectGetSpecificGeometryCall::ESpecifier geomType)
{
	return NULL;
}
